import time
import uuid
import string
import random

from psc_protocol.message_codes import MessageCode
from psc_protocol.request_factory import RequestFactory
from psc_protocol.util.enums import ReputationProperty, EffectiveReputationSource
from python_sensor_sim import PythonSensorSim

from psc_protocol.defense.request import (
    HelloRequest,
    DeviceStatusRequest,
    ReputationRequest,
    PolicyRequest,
    NetworkStatusRequest,
    ConfigurationRequest,
    UploadAPCFileRequest,
    UploadConferFileRequest,
    HashlistRefreshRequest,
    GetSoftwareUpgradeRequest,
    HashDeleteListRequest,
    SensorActionsRequestRequest,
    UninstallCodeRequest,
    BulkReputationRequest,
    BehaviorRequest,
    NetworkBehaviorRequest,
    APICallBehaviorRequest,
)


class MsgPackRequests:
    # pylint: disable=too-many-public-methods
    """
    Creates various message pack messages
    """

    def build_reputation_request(self, sensor: PythonSensorSim, apc_lookup: bool = False) -> ReputationRequest:
        """
        Create a reputation request
        :param sensor: Sensor object
        :param apc_lookup: Flag to indicate whether to perform apc_lookup
        :return: Reputation request
        """
        _time = int(time.time() * 1000)
        _sha256_hash = random.getrandbits(256)
        sha256_hash = hex(_sha256_hash)[2:]
        filename = str(uuid.uuid4())

        hash_request: ReputationRequest = RequestFactory.make_request(MessageCode.REQ_REPUTATION)
        hash_request.set(hash_request.Protocol.SHA256_HASH, sha256_hash)
        hash_request.set(hash_request.Protocol.MD5_HASH, None)
        hash_request.set(hash_request.Protocol.FILENAME, filename)
        hash_request.set(hash_request.Protocol.CURRENT_TIME, int(time.time() * 1000))
        hash_request.set(hash_request.Protocol.APC_LOOKUP, apc_lookup)
        self.configure_header(hash_request, sensor)
        return hash_request

    def build_policy_request(self, sensor: PythonSensorSim, protocol_version: int = 18) -> PolicyRequest:
        """
        Create a policy request
        :param sensor: Sensor object
        :param protocol_version: Version of protocol to use
        :return: policy request
        """
        policy_req = RequestFactory.make_request(MessageCode.REQ_POLICY)
        policy_req.set(policy_req.Header.VERSION, protocol_version)
        self.configure_header(policy_req, sensor)
        return policy_req

    def build_status_request(self, sensor: PythonSensorSim) -> DeviceStatusRequest:
        """
        Create a status request
        :param sensor: Sensor object
        :return: device status request
        """
        # TODO: Update to cover all field values
        # TODO: Make these parameters configurable with a data class
        local_ip_address = [random.randint(0, 255) for _ in range(4)]
        status_req = RequestFactory.make_request(MessageCode.REQ_DEVICE_STATUS)
        status_req.set(status_req.Header.LOCAL_IP_ADDRESS, {'IP': [2, bytes(local_ip_address)]})
        status_req.set(status_req.Protocol.PLATFORM_TYPE, 2)  # CLIENT x64
        status_req.set(status_req.Protocol.OS_MAJOR_VERSION, 1010)  # Windows 10
        status_req.set(status_req.Protocol.OS_MINOR_VERSION, 0)
        status_req.set(status_req.Protocol.MAINTENANCE_RELEASE, 0)
        status_req.set(status_req.Protocol.SENSOR_STATUS, [0])
        status_req.set(status_req.Protocol.HOSTNAME, sensor.device_host_name)
        status_req.set(status_req.Protocol.VERSION, '3.8.0.1612')
        status_req.set(status_req.Protocol.QUARANTINED, False)
        status_req.set(status_req.Protocol.LOGIN_USER_NAME, None)
        status_req.set(status_req.Protocol.OS_BUILD_NUMBER, None)
        self.configure_header(status_req, sensor)
        return status_req

    def build_hello_request(self, sensor: PythonSensorSim) -> HelloRequest:
        """
        Create a hello request
        :param sensor: Sensor object
        :return: hello request
        """
        hello_req = RequestFactory.make_request(MessageCode.REQ_HELLO)
        self.configure_header(hello_req, sensor)
        return hello_req

    def build_network_status_request(self, sensor: PythonSensorSim) -> NetworkStatusRequest:
        """
        Create a networ status request
        :param sensor: Sensor object
        :return: network status request
        """
        status_req = RequestFactory.make_request(MessageCode.REQ_NETWORK_STATUS)
        status_req.set(status_req.Protocol.CURRENT_TIME, int(time.time() * 1000))
        status_req.set(status_req.Protocol.INTERFACE_TYPE, random.choice([1, 2, 3]))
        # TODO: Add remaining missing fields
        self.configure_header(status_req, sensor)
        return status_req

    def build_configuration_request(self, sensor: PythonSensorSim) -> ConfigurationRequest:
        """
        Create a configuration request
        :param sensor: Sensor object
        :return: configuration request
        """
        config_req = RequestFactory.make_request(MessageCode.REQ_CONFIGURATION)
        self.configure_header(config_req, sensor)
        return config_req

    # /services/registration
    # /services/registrationcode
    # /services/deregister

    def build_upload_request(
        self, sensor: PythonSensorSim, file_hash: str, file_data: bytes, filename: str
    ) -> UploadAPCFileRequest:
        """
        Create a file upload request
        :param sensor: Sensor object
        :param file_hash: file hash
        :param file_data: data to upload
        :param filename: filename
        :return: file upload request
        """
        upload_req = RequestFactory.make_request(MessageCode.REQ_APC_UPLOAD)
        upload_req.set(upload_req.Protocol.CURRENT_TIME, int(time.time() * 1000))
        upload_req.set(upload_req.Protocol.SHA_HASH, file_hash)
        upload_req.set(upload_req.Protocol.FILENAME, filename)
        upload_req.set(upload_req.Protocol.FILE_SIZE, len(file_data))
        upload_req.set(upload_req.Protocol.COMPRESSED, 0)  # No Compression
        self.configure_header(upload_req, sensor)
        return upload_req

    def build_uploadconferfile_request(self, sensor: PythonSensorSim, filename: str) -> UploadConferFileRequest:
        """
        Create an upload confer file request
        :param sensor: Sensor object
        :param filename: filename
        :return: upload confer file request
        """
        upload_req = RequestFactory.make_request(MessageCode.REQ_UPLOAD_CONFER_FILE)
        upload_req.set(upload_req.Protocol.FILENAME, filename.encode())
        upload_req.set(upload_req.Protocol.TYPE, random.choice([0, 1, 2, 3, 4]))
        upload_req.set(upload_req.Protocol.AUTO, random.choice([True, False]))
        self.configure_header(upload_req, sensor)
        return upload_req

    # /services/mdmregistration

    def build_hash_list_refresh_v2_request(self, sensor: PythonSensorSim) -> HashlistRefreshRequest:
        """
        Create a hash list refresh request
        :param sensor: Sensor object
        :return: hasn list refresh request
        """
        hash_req = RequestFactory.make_request(MessageCode.REQ_HASHLIST_REFRESH)
        self.configure_header(hash_req, sensor)
        return hash_req

    def build_software_upgrade_request(self, sensor: PythonSensorSim) -> GetSoftwareUpgradeRequest:
        """
        Create a software upgrade request
        :param sensor: Sensor object
        :return: software upgrade request
        """
        upgrade_req = RequestFactory.make_request(MessageCode.REQ_GET_SOFTWARE_UPGRADE)
        self.configure_header(upgrade_req, sensor)
        return upgrade_req

    def build_get_hash_delete_list_request(self, sensor: PythonSensorSim) -> HashDeleteListRequest:
        """
        Create a hash delete list request
        :param sensor: Sensor object
        :return: hash delete list request
        """
        hash_req = RequestFactory.make_request(MessageCode.REQ_GET_HASH_DELETE_LIST)
        self.configure_header(hash_req, sensor)
        return hash_req

    # /services/alarms

    def build_sensor_action_request(self, sensor: PythonSensorSim) -> SensorActionsRequestRequest:
        """
        Create a sensor action request
        :param sensor: Sensor object
        :return: sensor action request
        """
        action_req = RequestFactory.make_request(MessageCode.REQ_SENSOR_ACTIONS_REQUEST)
        self.configure_header(action_req, sensor)
        return action_req

    # TODO: /services/sensorActionReport <-- requires request ID from /services/sensorActionRequest

    def build_zip_container_request(self, sensor: PythonSensorSim):
        """
        Create a zip container request
        :param sensor: Sensor object
        :return: zip container request
        """
        zip_req = RequestFactory.make_request(MessageCode.REQ_ZIP_CONTAINER)
        self.configure_header(zip_req, sensor)
        return zip_req

    def build_uninstallcode_request(self, sensor: PythonSensorSim) -> UninstallCodeRequest:
        """
        Create an uninstall request
        :param sensor: Sensor object
        :return: Uninstall request
        """
        uninstall_req = RequestFactory.make_request(MessageCode.REQ_UNINSTALL_CODE)
        uninstall_req.set(uninstall_req.Protocol.REQUEST_TYPE, random.choice([1, 2, 3]))
        self.configure_header(uninstall_req, sensor)
        return uninstall_req

    # def build_metadata_request(self, sensor: PythonSensorSim):
    #     meta_req = RequestFactory.make_request(MessageCode.REQ_METADATA)

    def build_bulk_reputation_v2_request(self, sensor: PythonSensorSim, reputations: list) -> BulkReputationRequest:
        """
        Create a bulk reputation request
        :param sensor: Sensor object
        :param reputations: list of reputations to build
        :return: bulk reputation request
        """
        rep_req = RequestFactory.make_request(MessageCode.REQ_BULK_REPUTATION)
        rep_req.set(rep_req.Protocol.COUNT, len(reputations))
        rep_req.set(rep_req.Protocol.COMPRESSION, False)
        rep_req.set(rep_req.Protocol.TYPE, 0)
        self.configure_header(rep_req, sensor)
        return rep_req

    def build_network_behavior_notification_request(self, sensor: PythonSensorSim) -> NetworkBehaviorRequest:
        """
        Create a network behavior notification request
        :param sensor: Sensor object
        :return: Network behavior request
        """
        beh_req = RequestFactory.make_request(MessageCode.REQ_NETWORK_BEHAVIOR)
        self._build_behavior_request(beh_req)
        self.configure_header(beh_req, sensor)
        return beh_req

    def build_data_access_behavior_notification(self, sensor: PythonSensorSim):
        """
        Create a data access behavior notification request
        :param sensor: Sensor object
        :return: behavior request
        """
        beh_req = RequestFactory.make_request(MessageCode.REQ_DATA_ACCESS_BEHAVIOR)
        beh_req.set(beh_req.BehaviorProtocol.DLP_FLAGS, [random.randint(0, 18)])
        beh_req.set(beh_req.BehaviorProtocol.FILE_NAME, self.generate_random_filename())
        beh_req.set(beh_req.BehaviorProtocol.FILE_FLAGS, [random.randint(0, 13)])
        self._build_behavior_request(beh_req)
        self.configure_header(beh_req, sensor)
        return beh_req

    def build_api_behavior_notification(self, sensor: PythonSensorSim) -> APICallBehaviorRequest:
        """
        Create api behavior notification
        :param sensor: Sensor object
        :return: Api call behavior request
        """
        beh_req = RequestFactory.make_request(MessageCode.REQ_API_CALL_BEHAVIOR)
        beh_req.set(beh_req.BehaviorProtocol.CLASS_TYPE, random.randint(4721, 4725))  # TODO: Cover all values
        beh_req.set(
            beh_req.BehaviorProtocol.CLASS_MASK, ''.join([random.choice(string.ascii_letters) for _ in range(10)])
        )
        beh_req.set(beh_req.BehaviorProtocol.TARGET_HASH, self.generate_random_sha256())
        beh_req.set(beh_req.BehaviorProtocol.TARGET_NAME, self.generate_random_filename())
        beh_req.set(beh_req.BehaviorProtocol.TARGET_PID, random.randint(10, 999999))
        beh_req.set(beh_req.BehaviorProtocol.TARGET_PROPERTY, ReputationProperty.NOT_LISTED.value)
        beh_req.set(beh_req.BehaviorProtocol.TARGET_USER_ID, random.randint(1, 999))
        beh_req.set(beh_req.BehaviorProtocol.TARGET_USERNAME, 'IronHide')
        beh_req.set(beh_req.BehaviorProtocol.TARGET_UNIQUE_PROCESS_ID, random.randint(0, 999999))
        beh_req.set(beh_req.BehaviorProtocol.TARGET_EFFECTIVE_REPUTATION_SOURCE, EffectiveReputationSource.CLOUD.value)
        beh_req.set(beh_req.BehaviorProtocol.TARGET_EFFECTIVE_REPUTATION, ReputationProperty.NOT_LISTED.value)
        self._build_behavior_request(beh_req)
        self.configure_header(beh_req, sensor)
        return beh_req

    def build_bulk_behavior_v2_request(self, sensor: PythonSensorSim, behaviors: list):
        """
        Create a bulk behavior request
        :param sensor: Sensor object
        :param behaviors: a list of behaviors to create
        :return: behavior request
        """
        beh_req = RequestFactory.make_request(MessageCode.REQ_BULK_BEHAVIOR)
        beh_req.set(beh_req.Protocol.COUNT, len(behaviors))
        beh_req.set(beh_req.Protocol.COMPRESSION, False)
        self.configure_header(beh_req, sensor)
        return beh_req

    def _build_behavior_request(self, beh_req: BehaviorRequest):
        """
        Create behavior request
        :param beh_req: behavior request
        """
        flow = [
            int(time.time() * 1000),
            [random.randint(0, 255) for _ in range(4)],
            [random.randint(0, 255) for _ in range(4)],
            random.randint(80, 60000),
            random.randint(10000, 60000),
            [random.randint(0, 31)],
            2,
            6,
        ]
        beh_req.set(beh_req.Protocol.PROCESS_HASH, self.generate_random_sha256())
        beh_req.set(beh_req.Protocol.PROCESS_NAME, self.generate_random_filename())
        beh_req.set(beh_req.Protocol.PROCESS_PROCESSID, random.randint(1, 99999))
        beh_req.set(beh_req.Protocol.DISPOSITION, [random.randint(0, 20)])
        beh_req.set(beh_req.Protocol.INFO, random.randint(0, 14))  # TODO: Cover all INFO values
        beh_req.set(beh_req.Protocol.CURRENT_TIME, int(time.time() * 1000))
        beh_req.set(beh_req.Protocol.FLOW, flow)
        beh_req.set(beh_req.Protocol.PROCESS_PROPERTY, ReputationProperty.NOT_LISTED.value)
        beh_req.set(beh_req.Protocol.UNDER_REVIEW, False)
        beh_req.set(beh_req.Protocol.PROCESS_TIME, int(time.time() * 1000))
        beh_req.set(beh_req.Protocol.PARENT_HASH, self.generate_random_sha256())
        beh_req.set(beh_req.Protocol.PARENT_NAME, self.generate_random_filename())
        beh_req.set(beh_req.Protocol.PARENT_ID, random.randint(100, 99999))
        beh_req.set(beh_req.Protocol.PARENT_PROPERTY, ReputationProperty.NOT_LISTED.value)
        beh_req.set(beh_req.Protocol.PROCESS_USERNAME, 'BumbleBee')
        beh_req.set(beh_req.Protocol.PROCESS_UNIQUE_PROCESSID, random.randint(10, 999999))
        beh_req.set(beh_req.Protocol.PARENT_USERNAME, 'OptimusPrime')
        beh_req.set(beh_req.Protocol.PARENT_UNIQUE_PROCESSID, random.randint(10, 999999))
        beh_req.set(beh_req.Protocol.GRANDPARENT_UNIQUE_PROCESSID, random.randint(10, 999999))
        beh_req.set(beh_req.Protocol.PROCESS_EFFECTIVE_REPUTATION, EffectiveReputationSource.CLOUD.value)
        beh_req.set(beh_req.Protocol.PARENT_EFFECTIVE_REPUTATION_SOURCE, EffectiveReputationSource.CLOUD.value)
        beh_req.set(beh_req.Protocol.PARENT_EFFECTIVE_REPUTATION, ReputationProperty.NOT_LISTED.value)

    @staticmethod
    def configure_header(request, sensor: PythonSensorSim):
        """
        Configures message header
        :param request: request to add header to
        :param sensor: Sensor object
        """
        request.set(request.Header.TIME_SENT, int(time.time() * 1000))
        request.set(request.Header.DEVICE_ID, sensor.guid)
        request.set(request.Header.DEVICE_TYPE, sensor.device_type.value)
        request.set(request.Header.LAST_UPDATE_TIME, sensor.last_update_time)
        request.set(request.Header.INFO, {'LastLiveQueryTime': sensor.last_live_query_time})
        request.set(request.Header.REGISTRATION_ID, sensor.registration_id)

    @staticmethod
    def generate_random_filename() -> str:
        """
        Generates a random filename
        :return: generated filename
        """
        _extensions = ['exe', 'msi', 'png', 'jpeg', 'dmg', 'zip', 'sh', 'xlsx', 'html', 'py', 'c', 'rb', '']
        _prefix = ''.join([random.choice(string.ascii_letters) for _ in range(10)])
        return f'{_prefix}_{random.randint(0, 99999999)}.{random.choice(_extensions)}'

    @staticmethod
    def generate_random_sha256() -> str:
        """
        Generates a random sha256 value
        :return: generated sha
        """
        _sha256_hash = random.getrandbits(256)
        return hex(_sha256_hash)[2:]
