import logging

from typing import Any, Tuple, List, Dict

from dataclasses import fields, Field

import proto_proto_pscr.cb_psc_event_pb2 as proto_message_types


class MessageDataClass:
    """
    Parent class inherited by dataclass instances. Provides logic for turning dataclass instance into the appropriate
    protobuf message type
    """

    PROTO_PRIMITIVE_TYPES = ['int', 'float', 'str', 'bool']
    PROTO_LIST_TYPES = [
        'RepeatedCompositeFieldContainer',
        'RepeatedScalarFieldContainer',
        'RepeatedCompositeContainer',
        'RepeatedScalarContainer',
    ]

    logger = logging.getLogger(__name__)

    @property
    def message_type(self):
        """
        The Protobuf message type that maps to the specific dataclass instance (e.g. CbEventMsg, CbProcessObject, etc).
        """
        return self._message_type  # noqa

    def create_message_object(self) -> Any:
        """
        Translates fields of data class into fields of a protobuf message object of supplied message type
        :return: Created message object
        """
        message_object = self.create_protobuf_message_of_type()
        for data_field in fields(self):
            if data_field.repr:  # Don't process _message_type field
                message_field_data_type_name, message_field_data_type_module = self.get_field_type_name_and_module(
                    message_object, data_field.name
                )
                field_data = getattr(self, data_field.name)

                # Handle blank or set timestamps
                if message_field_data_type_module == 'google.protobuf.timestamp_pb2':
                    self.add_timestamp_to_message(message_object, data_field.name, field_data)

                elif field_data:
                    # Handle nested event_messages
                    if message_field_data_type_module == 'cb_psc_event_pb2':
                        self.add_sub_message_to_message(
                            message_object, message_field_data_type_name, data_field, field_data
                        )

                    # Handle repeated Fields
                    elif message_field_data_type_name in self.PROTO_LIST_TYPES:
                        self.add_repeated_fields_to_message(
                            message_object, message_field_data_type_name, data_field, field_data
                        )

                    # Use setattr for primitive types
                    elif message_field_data_type_name in self.PROTO_PRIMITIVE_TYPES:
                        setattr(message_object, data_field.name, field_data)

        return message_object

    def add_sub_message_to_message(
        self, message_object: Any, message_field_data_type_name: str, data_field: Field, field_data: Any
    ):
        """
        Creates, if necessary sub message from field data and adds it to the specified message object at the specified
        field
        :param message_object: Message object to update
        :param message_field_data_type_name: The data type name of the field to update
        :param data_field: A Field object of data from the dataclass represented by self
        :param field_data: The data to set in the message field
        """
        field_data_type_name = type(field_data).__name__
        # Set sub message directly
        if field_data_type_name == message_field_data_type_name:
            self.add_sub_message(message_object, data_field.name, field_data)

        # Construct sub message from dictionary
        elif isinstance(field_data, dict):
            field_data = self.create_sub_message_from_dict(data_field, field_data)
            self.add_sub_message(message_object, data_field.name, field_data)

        # Construct sub message from dataclass
        elif field_data.message_type == message_field_data_type_name:
            sub_message = field_data.create_message_object()
            self.add_sub_message(message_object, data_field.name, sub_message)

    def add_repeated_fields_to_message(
        self, message_object: Any, message_field_data_type_name: str, data_field: Field, field_data
    ):
        """
        Adds list of values to a repeated message field
        :param message_object: Message object to update
        :param message_field_data_type_name: The data type name of the field to update
        :param data_field: A Field object of data from the dataclass represented by self
        :param field_data: The data to set in the message field
        """
        if 'Scalar' in message_field_data_type_name:
            self.add_scalar_list_values_to_message(message_object, data_field.name, field_data)
        else:
            # Process nested list of dict
            try:
                if all(isinstance(message, dict) for message in field_data):
                    field_data = self.create_sub_message_list(data_field, field_data)
                    self.add_composite_list_values_to_message(message_object, data_field.name, field_data)
                else:
                    list_sub_message_type = self.get_repeated_composite_field_message_type(
                        message_object, data_field.name
                    )
                    # Field data type matches expected data type. Set directly
                    if all((type(message).__name__ == list_sub_message_type) for message in field_data):
                        self.add_composite_list_values_to_message(message_object, data_field.name, field_data)
                    # Need to create sub message objects before setting
                    else:
                        message_object_list = [message.create_message_object() for message in field_data]
                        self.add_composite_list_values_to_message(message_object, data_field.name, message_object_list)
            except TypeError:  # Caller passed a single instance not wrapped in a list
                self.add_composite_list_values_to_message(message_object, data_field.name, [field_data])

    def get_repeated_composite_field_message_type(self, message: Any, field: str) -> str:
        """
        Gets the expected sub message type of a repeated composite field
        :param message: Message object to inspect
        :param field: The repeated composite field name
        :return: String of the data type name expected by the repeated composite field
        """
        command = f'message.{field}.add()'
        message_object = self._eval_protobuf_command(command, {'message': message})
        repeated_field_message_type = type(message_object).__name__
        # Reset composite field
        reset_command = f'del message.{field}[:]'
        self._exec_protobuf_command(reset_command, {'message': message})
        return repeated_field_message_type

    def create_protobuf_message_of_type(self) -> Any:
        """
        Creates a protobuf message object of type defined by self.message_type
        :return: The created protobuf message object
        """
        try:
            message_class = self._get_class_from_message_type()
            return message_class()
        except AttributeError:
            self.logger.warning(f'Could not instantiate message object of type: {self.message_type}.')
            return None

    def _get_class_from_message_type(self):
        """
        Gets leaf class from dot separated classpath
        :return: Class type at leaf of path described by self.message_type
        """
        types = self.message_type.split('.')
        current_class = proto_message_types
        for current_type in types:
            try:
                current_class = getattr(current_class, current_type)
            except AttributeError:
                self.logger.warning(f'Could not instantiate message object of type: {self.message_type}.')
                return None
        return current_class

    def create_sub_message_from_dict(self, data_field: Field, message_field_values: Dict) -> Any:
        """
        Generates protobuf message object of type specified in field_metadata['dataclass_type'] with specified values
        :param data_field: Dictionary of dataclass field metadata that specifies the dataclass type to create
        :param message_field_values: Values to set in created message object
        :return: Created protobuf message object
        """
        sub_message_data_type = data_field.metadata.get('dataclass_type')
        sub_message = None
        if sub_message_data_type:
            dataclass_object = sub_message_data_type(**message_field_values)
            sub_message = dataclass_object.create_message_object()
        else:
            self.logger.warning(
                f'Could not generate sub message object for field: {data_field.name} as there is'
                f'no sub message dataclass type mapping in the metadata for this field. Please raise'
                f'a ticket in PRDTEST to have this functionality added.'
            )
        return sub_message

    def create_sub_message_list(self, data_field: Field, field_dict: Dict) -> List[Any]:
        """
        Generates a list of protobuf event_messages of type specified in field_metadata['dataclass_type'] with specified
        values
        :param data_field: Dictionary of dataclass field metadata that specifies the dataclass type to create
        :param field_dict: List of value dictionaries to use for each generated message
        :return: A list of created protobuf message objects
        """
        sub_messages = []
        for message in field_dict:
            sub_messages.append(self.create_sub_message_from_dict(data_field, message))
        return sub_messages

    def get_field_type_name_and_module(self, message: Any, field_name: str) -> Tuple[str, str]:
        """
        Gets the field type name, and field type module name (if exists) of the supplied message field
        :param message: Message object for which to retrieve field type metadata
        :param field_name: Message field to retrieve type metadata from
        :return: Tuple of field type name (e.g. 'str', or 'CbProcessObject') and field type module name
        """
        field_type = None
        field_type_name = None
        field_type_module = None
        try:
            field_type = type(getattr(message, field_name))
            field_type_name = field_type.__name__
        except AttributeError:
            self.logger.warning(f'Could not determine field type for field named: {field_name}')

        if field_type_name:
            try:
                field_type_module = field_type.__module__
                if field_type_module == 'builtins':
                    field_type_module = None
            except AttributeError:
                self.logger.warning(
                    f'Could not determine module name for field: {field_name} - type: ' f'{field_type_name}'
                )
        return field_type_name, field_type_module

    def add_timestamp_to_message(self, message: Any, timestamp_field_name: str, timestamp: str = None):
        """
        Sets a timestamp field on a protobuf message object. If timestamp is None, the current time will be set in the
        specified field
        :param message: Message object to update
        :param timestamp_field_name: Message field to update
        :param timestamp: Optional timestamp value as a string (e.g. "1970-01-01T00:00:00Z")
        """
        if timestamp:
            command = f'message.{timestamp_field_name}.FromJsonString(timestamp)'
            self._exec_protobuf_command(command, {'message': message, 'timestamp': timestamp})
        else:
            command = f'message.{timestamp_field_name}.GetCurrentTime()'
            self._exec_protobuf_command(command, {'message': message})

    def add_scalar_list_values_to_message(self, message: Any, field_name: str, values: List[Any]):
        """
        Updates the specified message at the specified field with the specified list of scalar values
        (e.g. str, int, etc.)
        :param message: Message object to update
        :param field_name: Repeated field name
        :param values: List of values to set in the field
        """
        command = f'message.{field_name}[:] = values'
        self._exec_protobuf_command(command, {'message': message, 'values': values})

    def add_composite_list_values_to_message(self, message: Any, field_name: str, values: List[Any]):
        """
        Updates the specified message at the specified field with the specified list of composite values
        (e.g. message objects)
        :param message: Message object to update
        :param field_name: Repeated field name
        :param values: List of values to set in the field
        """
        command = f'message.{field_name}.extend([value])'
        for value in values:
            self._exec_protobuf_command(command, {'message': message, 'value': value})

    def add_sub_message(self, message: Any, field_name: str, sub_message: Any):
        """
        Performs a CopyFrom operation on the specified message object/field to set the specified sub_message value
        :param message: Message object to update
        :param field_name: Name of sub message field
        :param sub_message: Sub message to set in the field
        """
        command = f'message.{field_name}.CopyFrom(sub_message)'
        self._exec_protobuf_command(command, {'message': message, 'sub_message': sub_message})

    @staticmethod
    def _exec_protobuf_command(command: str, arguments: Dict[str, Any]):
        """
        Executes the supplied command with the supplied arguments as local variables
        :param command: Command to execute
        :param arguments: Arguments for command
        """
        exec(command, {}, arguments)

    @staticmethod
    def _eval_protobuf_command(command: str, arguments: Dict[str, Any]) -> Any:
        """
        Executes the supplied command with the supplied arguments as local variables and returns the result
        :param command: Command to execute
        :param arguments: Arguments for command
        :return: The result of command evaluation
        """
        return eval(command, {}, arguments)
