import time
import os
import logging

from typing import Dict

from zipfile import ZipFile

from proto_proto_pscr.cb_psc_event_pb2 import CbEventBatchMsg, CbEventMsg, CbEventProtocolVersion

from psc_server_support.api.device_services import DeviceServices

from sensor_comms.protobuf_requests.event_messages.event.event import Event
from sensor_comms.protobuf_requests.event_messages.event.event_header import EventHeader
from sensor_comms.protobuf_requests.event_messages.objects.process_object import ProcessObject
from sensor_comms.protobuf_requests.event_messages.objects.system_object import SystemObject
from sensor_comms.protobuf_requests.event_messages.operations.network_operation import NetworkOperation
from sensor_comms.protobuf_requests.event_messages.operations.process_operation import ProcessOperation

from sensor_comms.protobuf_requests.utilities.utilities import raise_on_failure_code


class EventMessageBuilder:
    """
    Builds protobuf events and sends them to the backend
    """

    def __init__(self):
        self.logger = logging.getLogger(__name__)
        self._event = None
        self._events = self._reset_batch()

    @property
    def event(self) -> Event:
        """
        Event property
        """
        return self._event

    @property
    def events(self):
        """
        Event batch property
        """
        return self._events

    def create_new_event(self, **kwargs) -> 'EventMessageBuilder':
        """
        Creates new event and adds it to the builder
        :param kwargs: A dictionary of values to set in the fields of the event message. See fields of
        :py:class:`~protobuf_requests.event_messages.event.Event`
        :return: self
        """
        self._event = Event(**kwargs)
        return self

    def add_header(self, **kwargs):
        """
        Adds a CbEventHeader to the current event with the values specified in the keyword arguments
        :param kwargs: A dictionary of values to set in the fields of the event header. See fields of
        :py:class:`protobuf_requests.event_messages.event.EventHeader`
        """
        self._event.header = EventHeader(**kwargs).create_message_object()

    def add_actor(self, **kwargs):
        """
        Adds an actor to current message
        :param kwargs: A dictionary of ProcessObject field values. See
        :py:class:`~protobuf_requests.event_messages.objects.process_object.ProcessObject` for available fields.
        Nested message fields can be set using a dictionary
        """
        self._event.actor = ProcessObject(**kwargs).create_message_object()

    def add_parent_actor(self, **kwargs):
        """
        Adds a parent actor to current message
        :param kwargs: A dictionary of ProcessObject field values. See
        :py:class:`~protobuf_requests.event_messages.objects.process_object.ProcessObject` for available fields.
        Nested message fields can be set using a dictionary
        """
        self._event.parent_actor = ProcessObject(**kwargs).create_message_object()

    def add_grand_parent_actor(self, **kwargs):
        """
        Adds a grandparent actor to current message
        :param kwargs: A dictionary of ProcessObject field values. See
        :py:class:`~protobuf_requests.event_messages.objects.process_object.ProcessObject` for available fields.
        Nested message fields can be set using a dictionary
        """
        self._event.grandparent_actor = ProcessObject(**kwargs).create_message_object()

    def add_operation(self, operation_type: str, **kwargs):
        """
        Adds specified operation type to the current event message
        :param operation_type: String of event type
        :param kwargs: A dictionary of values to set in the operation message
        """
        if operation_type == 'netconn_op':
            self._event.netconn_op = NetworkOperation(**kwargs).create_message_object()
        elif operation_type == 'process_op':
            self._event.process_op = ProcessOperation(**kwargs).create_message_object()
        else:
            raise NotImplementedError

    def add_system_object(self, **kwargs):
        """
        Adds a system object to the current event message
        :param kwargs: A dictionary of values to set in the fields of the system object. See fields of
        :py:class:`~protobuf_requests.event_messages.objects.system_object.SystemObject`
        """
        self._event.system = SystemObject(**kwargs).create_message_object()

    def set_storage_class(self, storage_class: str):
        """
        Sets the storage class on the current mesasge
        :param storage_class: String of storage class. Default is 'default'
        """
        self._event.storage_class = storage_class

    def add_exec_parent_actor(self, **kwargs):
        """
        Adds an exec parent process/actor to current message
        :param kwargs: A dictionary of ProcessObject field values. See
        :py:class:`~protobuf_requests.event_messages.objects.process_object.ProcessObject` for available fields.
        Nested message fields can be set using a dictionary
        """
        self._event.exec_parent_actor = ProcessObject(**kwargs).create_message_object()

    def add_exec_grandparent_actor(self, **kwargs):
        """
        Adds an exec grandparent process/actor to current message
        :param kwargs: A dictionary of ProcessObject field values. See
        :py:class:`~protobuf_requests.event_messages.objects.process_object.ProcessObject` for available fields.
        Nested message fields can be set using a dictionary
        """
        self._event.exec_grandparent_actor = ProcessObject(**kwargs).create_message_object()

    def build(self) -> CbEventMsg:
        """
        Creates CbEventMsg from current Event object
        :return: Created protobuf message object
        """
        built_event = self.event.create_message_object()
        self._events.events.extend([built_event])
        self._event = None
        return built_event

    def send_messages(self, header: Dict, url: str = None):
        """
        Sends current batch of messages to CBC backend
        :param header: Dictionary of protobuf header values
        :param url: Optional string of devices URL
        """
        if self._event:
            self.build()
        event_file = self._flush_batch_message_to_file()
        zipped_event_file = self._zip_events(event_file)
        self._send_events(header, zipped_event_file, url)
        self._remove_file(zipped_event_file)

    @staticmethod
    def _reset_batch():
        """
        Creates new message batch and set version
        :return: CbEventBatchMsg
        """
        batch = CbEventBatchMsg()
        batch.version = CbEventProtocolVersion.Value('PROTOCOL_VERSION_CURRENT')
        return batch

    def _flush_batch_message_to_file(self) -> str:
        """
        Serializes current batch of event messages to a file
        :return: String of serialized events file path
        """
        filename = 'test_event_{}'.format(str(int(time.time() * 1000000)))
        with open(filename, 'wb') as file:
            file.write(self.events.SerializeToString())
        self._events = self._reset_batch()
        return filename

    @staticmethod
    def _send_events(protobuf_headers: Dict, zipped_events: str, url: str = None):
        """
        Sends the passed in zipped event to the CBC backend
        :param protobuf_headers: A Dictionary of header values required by message
        :param zipped_events: String file path to zipped events
        :param url: Optional, url for devices
        """
        if not url:
            url = os.environ.get('DEVICES', '')
        device_services = DeviceServices(devices_url=url)
        response = device_services.post_threat_hunter_event(protobuf_headers, zipped_events)
        raise_on_failure_code(response)

    def _zip_events(self, event_file: str) -> str:
        """
        Zips the specified file and returns path to zipped file
        :param event_file: String of path to file to be zipped
        :return: String of zipped filename
        """
        zip_filename = 'test_event_batch_{}.zip'.format(str(int(time.time() * 1000000)))
        with ZipFile(zip_filename, 'w') as zipfile:
            zipfile.write(event_file)
        self._remove_file(event_file)
        return zip_filename

    def _remove_file(self, filename: str) -> bool:
        """
        Attempts to delete the specified filename
        :param filename: String of file path
        :return: Bool, True if deleted successfully, otherwise False
        """
        try:
            os.remove(filename)
            return True
        except OSError:
            self.logger.exception('Error removing file %s', filename)
            return False
