import hashlib
import random
import uuid
import json

from requests import Response

from psc_server_support.custom_exceptions import ResponseNotSuccessException


def raise_on_failure_code(response: Response, error_message: str = None) -> None:
    """
    Verifies that a response has an ok status code 2xx or 3xx.

    :param response: Requests Response to check
    :param error_message: (optional) Option message to add to the exception
    :raises ResponseNotSuccessException: On Non success
    """
    message = 'Expected successful 2xx or 3xx response status code.'
    if error_message:
        message = '{} {}'.format(error_message, message)
    if not response.ok:
        raise ResponseNotSuccessException(
            '{} Actual: {}, content: {}'.format(message, response.status_code, response.text)
        )

    # check success field if it is present in the response
    success = None
    try:
        json.loads(response.text)
        success = response.json().get('success')
    except ValueError:
        pass
    if success is not None:
        if not success:
            raise ResponseNotSuccessException(
                'Expected success field to be true in the response. Actual: {}'.format(success)
            )


def create_default_pid() -> int:
    """
    Creates a random int between 1 and 10000000 for use as a process pid
    :return: Random integer
    """
    return random.randint(1, 10000000)


def create_default_thread_id() -> int:
    """
    Creates a random int between 1 and 100000 for use as a thread id
    :return: Random integer
    """
    return random.randint(1, 100000)


def create_default_uuid() -> str:
    """
    Creates a random UUID and returns it as a string
    :return: UUID as string
    """
    return str(uuid.uuid4())


def create_default_md5_hash():
    """
    Generates a random md5 hash
    :return: String of a randomly generated md5 hash
    """
    random_bits = str(random.getrandbits(256)).encode('utf-8')
    return hashlib.md5(random_bits, usedforsecurity=False).hexdigest()


def create_default_sha256_hash():
    """
    Generates a random sha256 hash
    :return: String of a randomly generated sha256 hash
    """
    random_bits = str(random.getrandbits(256)).encode('utf-8')
    return hashlib.sha256(random_bits).hexdigest()
