import io
import os
import hashlib

from csp_cbc_auth.interfaces.cbc import CBC
from psc_server_support.api.api_factory import ApiFactory
from python_sensor_sim import PythonSensorSim
from sensor_comms.msgpack_requests.msg_pack_requests import MsgPackRequests
from locust import TaskSet, task


class DeviceServicesTaskSet(TaskSet):
    def __init__(self, parent):
        super().__init__(parent)
        cbc = CBC.auth_oauth_client_credentials()
        self.org_user = ApiFactory(cbc.cbc_session)
        self.sensor = self.init_sensor()
        self.msg_pack = MsgPackRequests()

    def init_sensor(self) -> PythonSensorSim:
        # TODO: Make this more dynamic and configurable
        return PythonSensorSim(org_user=self.org_user)

    @task
    def task_bulk_reputation_v2_message(self):
        route = '/services/bulkreputationV2'
        reputations = [self.msg_pack.build_reputation_request(sensor=self.sensor) for _ in range(10)]
        raw_data = io.BytesIO()

        for _rep in reputations:
            _pack = _rep.pack()
            raw_data.write(len(_pack).to_bytes(4, 'big'))
            raw_data.write(_pack)

        request = self.msg_pack.build_bulk_reputation_v2_request(sensor=self.sensor, reputations=reputations)
        self.send_mp_request_stream(route, raw_data, request)

    @task
    def task_bulk_behavior_v2_message(self):
        route = '/services/bulkbehaviorV2'
        raw_data = io.BytesIO()
        behaviors = [
            self.msg_pack.build_api_behavior_notification(sensor=self.sensor),
            self.msg_pack.build_network_behavior_notification_request(sensor=self.sensor),
            self.msg_pack.build_data_access_behavior_notification(sensor=self.sensor),
            self.msg_pack.build_api_behavior_notification(sensor=self.sensor),
        ]

        for behavior in behaviors:
            _pack = behavior.pack()
            raw_data.write(len(_pack).to_bytes(4, 'big'))
            raw_data.write(_pack)

        request = self.msg_pack.build_bulk_behavior_v2_request(sensor=self.sensor, behaviors=behaviors)
        self.send_mp_request_stream(route, raw_data, request)

    @task
    def task_reputation_message(self):
        route = '/services/reputation'
        request = self.msg_pack.build_reputation_request(sensor=self.sensor)
        self.send_mp_request(route, request)

    @task
    def task_behavior_message(self):
        route = '/services/behavior'
        request = self.msg_pack.build_api_behavior_notification(sensor=self.sensor)
        self.send_mp_request(route, request)

    @task
    def task_request_policy_message(self):
        route = '/services/policyV2'
        request = self.msg_pack.build_policy_request(sensor=self.sensor)
        self.send_mp_request(route, request)

    @task
    def task_send_hello_message(self):
        route = '/services/hello'
        request = self.msg_pack.build_hello_request(sensor=self.sensor)
        self.send_mp_request(route, request)

    @task
    def task_send_status_message(self):
        route = '/services/status'
        request = self.msg_pack.build_status_request(sensor=self.sensor)
        self.send_mp_request(route, request)

    @task
    def task_network_status_message(self):
        route = '/services/networkstatus'
        request = self.msg_pack.build_network_status_request(sensor=self.sensor)
        self.send_mp_request(route, request)

    @task
    def task_configuration_message(self):
        route = '/services/configuration'
        request = self.msg_pack.build_configuration_request(sensor=self.sensor)
        self.send_mp_request(route, request)

    @task
    def task_get_hash_delete_list_message(self):
        route = '/services/gethashdeletelist'
        request = self.msg_pack.build_get_hash_delete_list_request(sensor=self.sensor)
        self.send_mp_request(route, request)

    @task
    def task_uninstall_code_message(self):
        route = '/services/uninstallcode'
        request = self.msg_pack.build_uninstallcode_request(sensor=self.sensor)
        self.send_mp_request(route, request)

    @task
    def task_upload_message(self):
        route = '/services/upload'
        _filename = self.msg_pack.generate_random_filename()
        _data = io.BytesIO()
        _data.write(os.urandom(1024))

        _data.seek(0)
        _hash = hashlib.sha256(_data.read()).hexdigest()

        _data.seek(0)
        request = self.msg_pack.build_upload_request(
            sensor=self.sensor, file_hash=_hash, file_data=_data.read(), filename=_filename
        )

        self.send_mp_request_upload(route, _filename, _data, request)

    @task
    def task_uploadconfer_message(self):
        route = '/services/uploadconferfile'
        _filename = self.msg_pack.generate_random_filename()
        _data = io.BytesIO()
        _data.write(os.urandom(1024))

        _data.seek(0)
        _hash = hashlib.sha256(_data.read()).hexdigest()

        _data.seek(0)
        request = self.msg_pack.build_uploadconferfile_request(sensor=self.sensor, filename=_filename)

        self.send_mp_request_upload(route, _filename, _data, request)

    def send_mp_request(self, route: str, request):
        files = {'message': ('private', request.pack(), 'application/octet-stream')}
        self.client.post(route, files=files)

    def send_mp_request_upload(self, route: str, filename: str, data: io.BytesIO, request):
        data.seek(0)
        files = {
            'message': ('private', request.pack(), 'application/octet-stream'),
            'file': (filename, data.read(), 'application/octet-stream; charset=UTF-8;'),
        }
        self.client.post(route, files=files)

    def send_mp_request_stream(self, route: str, data: io.BytesIO, request):
        data.seek(0)
        files = {
            'message': ('private', request.pack(), 'application/octet-stream'),
            'datastream': ('private', data.read(), 'application/octet-stream; charset=UTF-8;'),
        }
        self.client.post(route, files=files)
