from sensor_comms.protobuf_requests.message_builders.event_message_builder import EventMessageBuilder
from sensor_comms.protobuf_requests.event_messages.event.rule_policy_guid import RulePolicyGuid
from sensor_comms.protobuf_requests.event_messages.objects.file_object import FileObject

from tests.test_data.event_data_examples import XDR_EVENT_DATA

from unittest import TestCase


class TestBuilder(TestCase):
    def test_create_default_message(self):
        """
        Create event with default arguments
        :return:
        """
        event = EventMessageBuilder().create_new_event().build()
        assert event.header.event_type == 3
        assert event.actor.main_module.path == 'C:\\WINDOWS\\system32\\services.exe'

    def test_create_from_nested_dict(self):
        """
        Create a complex event with nested sub messages from a dictionary
        """
        builder = EventMessageBuilder()
        event = builder.create_new_event(**XDR_EVENT_DATA).build()
        assert event.header.event_type == 4
        assert event.netconn_op.adapter_info.friendly_name == 'Ethernet0'

    def test_add_event_operation(self):
        """
        Add a supported operation to an event with default field values
        """
        event_builder = EventMessageBuilder().create_new_event()
        event_builder.add_operation('netconn_op')
        event = event_builder.build()
        assert event.netconn_op.action == 512

    def test_add_not_implemented_operation(self):
        """
        Add an unsupported operation to an event
        """
        event_builder = EventMessageBuilder().create_new_event()
        with self.assertRaises(NotImplementedError):
            event_builder.add_operation('will_never_exist_op')

    def test_multiple_events(self):
        """
        Create multiple events in by builder
        """
        builder = EventMessageBuilder()
        builder.create_new_event()
        event_msg = builder.build()
        builder.create_new_event().add_header(event_type=3)
        event2 = builder.build()
        assert event_msg != event2

    def test_build_event_incrementally(self):
        """
        Tests that an event can be built by manually building sub messges from their dataclasses. In this test
        the caller doesn't create the message object (i.e. CbRulePolicyGuid) from the data class. This test asserts
        that the message event generation logic will do that for you.
        """
        trigger_event_description = 'this is an event'
        builder = EventMessageBuilder()
        builder.create_new_event()
        triggers = [RulePolicyGuid(event_description=trigger_event_description)]
        builder.add_header(triggers=triggers)
        event = builder.build()
        assert event.header.triggers[0].event_description == trigger_event_description

    def test_build_event_incrementally_with_constructed_sub_messages(self):
        """
        Tests that an event can be built by manually building sub messges from their dataclasses. In this test
        the caller DOES create the message object (i.e. CbRulePolicyGuid) from the data class.
        """
        trigger_event_description = 'this is an event'
        builder = EventMessageBuilder()
        builder.create_new_event()
        triggers = [RulePolicyGuid(event_description=trigger_event_description).create_message_object()]
        builder.add_header(triggers=triggers)
        event = builder.build()
        assert event.header.triggers[0].event_description == trigger_event_description

    def test_build_event_message_with_sub_message_as_dataclass(self):
        parent_actor_file_object_tags = ['tag1']
        builder = EventMessageBuilder()
        builder.create_new_event()
        file_object = FileObject(tags=parent_actor_file_object_tags)
        builder.add_parent_actor(main_module=file_object)
        event = builder.build()
        assert event.parent_actor.main_module.tags == parent_actor_file_object_tags
