import importlib
import pkgutil
import typing
import pytest

from typing_inspect import get_args, get_origin

from dataclasses import fields

from proto_proto_pscr import cb_psc_event_pb2

import sensor_comms.protobuf_requests.event_messages as messages


def get_message_types():
    """
    Returns a list of all dataclasses in the event_messages module
    """
    path_list = []
    spec_list = []
    msg_classes = []
    for importer, modname, is_package in pkgutil.walk_packages(messages.__path__):
        import_path = f"{messages.__name__}.{modname}"
        if is_package:
            spec = pkgutil._get_spec(importer, modname)
            importlib._bootstrap._load(spec)
            spec_list.append(spec)
        else:
            path_list.append(import_path)
    for path in path_list:
        module = importlib.import_module(path)
        module_classes = module.__dict__
        for cls in module_classes:
            if isinstance(module_classes[cls], type) and module_classes[cls].__module__ == module.__name__:
                msg_classes.append(module_classes[cls])
    return msg_classes


class TestMessageTypes:
    @pytest.mark.parametrize('message_type', get_message_types())
    def test_dataclasses_have_message_type_attribute(self, message_type):
        """
        Asserts that each message type has a properly configured '_message_type' field
        """
        dataclass_fields = fields(message_type)
        has_message_type_field = False
        for field in dataclass_fields:
            if field.name == '_message_type':
                has_message_type_field = True
                assert not field.repr, 'The _message_type field should set "repr" to False'
                assert not field.init, 'The _message_type field should set "init" to False'
        assert has_message_type_field, 'All dataclasses must specify a "_message_type" field'

    @pytest.mark.parametrize('message_type', get_message_types())
    def test_dataclass_fields_map_to_message_fields(self, message_type):
        """
        Asserts that each message type dataclass has fields for each field of the corresponding protobuf message
        """
        dataclass_fields = fields(message_type)
        dataclass_instance = message_type()
        message_instance = dataclass_instance.create_protobuf_message_of_type()
        message_field_names = [field.name for field in message_instance.DESCRIPTOR.fields]
        for dataclass_field in dataclass_fields:
            if dataclass_field.name == 'record_class' and 'class' in message_field_names:
                message_field_names.remove('class')
            elif dataclass_field.name in message_field_names:
                message_field_names.remove(dataclass_field.name)

        assert len(message_field_names) == 0, (
            f'The following fields are missing from the dataclass'
            f' {message_field_names}. The protobuf schema may have been updated.'
        )

    @pytest.mark.parametrize('message_type', get_message_types())
    def test_metadata_mapping(self, message_type):
        """
        Asserts that datclass fields that map to protobuf message types specify a dataclass mapping for that type
        """
        cb_message_types = dir(cb_psc_event_pb2)
        dataclass_fields = fields(message_type)
        for dataclass_field in dataclass_fields:
            if get_origin(dataclass_field.type) == list:
                field_type_name = get_args(dataclass_field.type)[0].__name__
            else:
                field_type_name = dataclass_field.type.__name__

            if field_type_name in cb_message_types:
                metadata = dataclass_field.metadata
                mapped_dataclass = metadata.get('dataclass_type')
                assert mapped_dataclass, (
                    f'Field {dataclass_field.name} is of type {field_type_name}, '
                    f'but the field definition does not specify a dataclass mapping'
                )
